package Data::CompactReadonly::V0::Scalar::NegativeShort;
our $VERSION = '0.0.5';

use warnings;
use strict;
use base qw(
    Data::CompactReadonly::V0::NegativeScalar
    Data::CompactReadonly::V0::Scalar::Short
);

1;
